/******************************************************************************
  Copyright (c) 2006 by Turku PET Centre

  File: imgfilter.c
  
  Purpose: Gaussian filter.
  
  Date: 2005-08-08

  Author: Calle Laakkonen, Kaisa Sederholm

  Version:

  2006-08-08 Kaisa Sederholm

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
/*****************************************************************************/
#include "libtpcimgiov.h"
#include "img.h"
#include "include/imgfilter.h"
/*****************************************************************************/

/*****************************************************************************/
/** \internal
  \brief Allocate memory for a float matrix
  
  The matrix will be in form matrix[h][w]

  \param w width of matrix
  \param h height of matrix
  \return the allocated memory. NULL on error
  */
static float **mallocMatrix(float w,float h)
{
    int y;
    float **matrix = malloc(h * sizeof(float*));
    if(!matrix)
        return NULL;
    for(y=0;y<h;y++) {
        /* Allocate a row */
        matrix[y] = malloc(w*sizeof(float));

        if(!matrix[y]) {
            while(y>0)
                free(matrix[--y]);
            free(matrix);
            return NULL;
        }
    }
    return matrix;
}

/** \brief Make a gaussian convolution kernel
  \param size kernel size
  \return the convolution kernel. NULL on error
*/
float **imgGaussKernel(int size) {
    float **kernel;
    int x,y;
    float stdev;
    kernel = mallocMatrix(size,size);
    if(!kernel)
        return NULL;

    stdev = size/7.0;
    stdev = stdev*stdev;

    for(x=0;x<size;x++) {
        float mx = x - size/2;
        for(y=0;y<size;y++) {
            float my = y - size/2;
            kernel[x][y] = (1.0/(2*M_PI*stdev))*pow(M_E,-((mx*mx)+(my*my))/(2.0*stdev));
        }
    }
    return kernel;
}

/** \brief Free a convolution kernel
  \param kernel the convolution kernel
  \param size size of the kernel
  */
void imgFreeKernel(float **kernel, int size) {
    int r;
    for(r=0;r<size;r++)
        free(kernel[r]);
    free(kernel);
}

/** \brief Perform a convolution operation on float data

  \param data data in format float[y][x][frame]
  \param buffer temporary buffer to use. Size must be float[size][data width+size]
  \param frame frame to extract from data
  \param width width of the image
  \param height height of the image
  \param kernel convolution kernel
  \param size convolution kernel size
  */
void imgConvoluteData(float ***data, float **buffer, int frame, int width, int height, float **kernel, int size)
{
    const int size2 = size/2;
    int x,y;

    /* Fill the initial buffer */
    for(y=0;y<size2;y++) {
        /* Pixels above the border */
        for(x=0;x<width+size;x++)
            buffer[y][x] = 0;

        /* Padding for the borders */
        for(x=0;x<size2;x++) {
            buffer[y+size2][x] = 0;
            buffer[y+size2][x+width+size2] = 0;
        }

        /* Rows from the image */
        for(x=0;x<width;x++)
            buffer[y+size2][x+size2] = data[y][x][frame];
    }

    /* Filter */
    for(y=0;y<height;y++) {
        float *tmp;
        /* Filter a row */
        for(x=0;x<width;x++) {
            float filtered = 0;
            int kx,ky;
            for(kx=0;kx<size;kx++) {
                for(ky=0;ky<size;ky++) {
                    filtered += buffer[ky][x+kx] * kernel[kx][ky];
                }
            }
            data[y][x][frame] = filtered;
        }

        /* Shift rows */
        tmp = buffer[size-1];
        buffer[size-1] = buffer[0];
        for(x=0;x<size-2;x++)
            buffer[x] = buffer[x+1];
        buffer[size-2] = tmp;

        /* Get next row from image or blank */
        if(y+size2<height) {
            for(x=0;x<width;x++)
                buffer[size-1][x+size2] = data[y+size2][x][frame];
        } else {
            for(x=0;x<width;x++)
                buffer[size-1][x+size2] = 0;
        }
    }
}

/** \brief Perform a convolution operation on image

  The bordering size/2 pixels of the image do not containg valid data.
  imgConvolute is quite memory efficient.
  Only (img->dimx+size)*size*sizeof(float) bytes are consumed by a
  temporary buffer. When processing multiple planes/frames/images with
  equal width and kernel size, efficiency can be improved further by
  allocating the temporary buffer yourself and calling imgConvoluteData
  directly.

  \param img image
  \param frame frame to perform operation on
  \param plane plane to perform operation on
  \param kernel convolution kernel
  \param size convolution kernel size
  \pre size must be smaller than img->dimx and img->dimy
  \return nonzero in case of (malloc) error
*/
int imgConvolute(IMG *img, int frame, int plane, float **kernel, int size)
{
    float **buffer;
    int y;

    /* Allocate memory for temporary buffer */
    buffer = mallocMatrix(img->dimx+size,size);
    if(!buffer)
        return 1;

    /* Convolute */
    imgConvoluteData(img->m[plane],buffer,frame,img->dimx, img->dimy, kernel,size);

    /* Free buffer */
    for(y=0;y<size;++y)
        free(buffer[y]);
    free(buffer);
    return 0;
}
